/** @file sdl_socket.cpp

	Skelton for retropc emulator
	SDL edition

	@author Sasaji
	@date   2013.12.06

	@brief [ sdl socket ]

	@note
	This code is based on win32_socket.cpp of the Common Source Code Project.
	Author : Takeda.Toshiya
	Date   : 2006.08.27 -
*/

#include "sdl_emu.h"
#include "../../vm/vm.h"
#include "../../vm/device.h"
#include "../../cmutex.h"

#define USE_SDL_EVENT

#ifndef USE_SDL_EVENT
static EMU *emu;
// for callback timer
typedef struct {
	int ch;
	SDL_TimerID id;
} st_timer;
// callback
static Uint32 on_connected(Uint32 interval, void *param);
static Uint32 on_disconnected(Uint32 interval, void *param);
static Uint32 on_writeable(Uint32 interval, void *param);
static Uint32 on_readable(Uint32 interval, void *param);
#endif

#ifdef USE_SOCKET
#ifdef USE_GTK
static void OnUserSocket(GtkWidget *widget, gint data, gpointer user_data);
#endif

void EMU_OSD::EMU_SOCKET()
{
	// init sockets
	for(int i = 0; i < SOCKET_MAX; i++) {
//		socs[i].soc = INVALID_SOCKET;
		socs[i].device = NULL;
		socs[i].tcp_socket = NULL;
		socs[i].udp_socket = NULL;
		socs[i].is_tcp = false;
		socs[i].is_server = false;
		socs[i].socket_delay = 0;
		socs[i].recv_buffer = NULL;
		socs[i].recv_r_ptr = 0;
		socs[i].recv_w_ptr = 0;
		socs[i].mux = NULL;
	}
	socket_set = NULL;
	socket_nums = 0;

#ifdef USE_GTK
	// create new user defined signal
	g_signal_new("user-socket",
		GTK_TYPE_WIDGET,
		G_SIGNAL_RUN_LAST,
		0,
		NULL, NULL,
		g_cclosure_marshal_VOID__INT,
		G_TYPE_NONE, 1, G_TYPE_INT);
#endif
}

void EMU_OSD::initialize_socket()
{
#ifndef USE_SDL_EVENT
	emu = this;
#endif

	// init sdl net
	if(SDLNet_Init()==-1) {
		logging->out_logf(LOG_ERROR, _T("SDLNet_Init: %s"), SDLNet_GetError());
	}
	socket_set = SDLNet_AllocSocketSet(SOCKET_MAX);
	if(!socket_set) {
		logging->out_logf(LOG_ERROR, _T("SDLNet_AllocSocketSet: %s"), SDLNet_GetError());
	}

#ifdef USE_GTK
	// connect signal
	g_signal_connect(G_OBJECT(window), "user-socket",
		G_CALLBACK(OnUserSocket), (gpointer)this);
#endif
}

void EMU_OSD::release_socket()
{
	// release sockets
	for(int i = 0; i < SOCKET_MAX; i++) {
//		if(socs[i].soc != INVALID_SOCKET) {
//			shutdown(socs[i].soc, 2);
//			closesocket(socs[i].soc);
//		}
		SDLNet_TCP_Close(socs[i].tcp_socket);
		SDLNet_UDP_Close(socs[i].udp_socket);
		socs[i].tcp_socket = NULL;
		socs[i].udp_socket = NULL;

		delete[] socs[i].recv_buffer;
		delete socs[i].mux;
	}

	SDLNet_FreeSocketSet(socket_set);
	socket_set = NULL;

	// release sdl net
	SDLNet_Quit();
}

void EMU_OSD::socket_connected(int ch)
{
	// network is connected
	logging->out_debugf(_T("connected. ch=%d"), ch);
	if (socs[ch].device) socs[ch].device->network_connected(ch);
//	else vm->network_connected(ch);
}

void EMU_OSD::socket_disconnected(int ch)
{
	// network is disconnected
//	if(!socs[ch].socket_delay) {
//		socs[ch].socket_delay = 1;//56;
//	}
	logging->out_debugf(_T("disconnected. ch=%d"), ch);
}

/// @note called by main thread
void EMU_OSD::socket_writeable(int ch)
{
	// network is writeable
	logging->out_debugf(_T("writeable. ch=%d"), ch);
	if (socs[ch].device) socs[ch].device->network_writeable(ch);
//	else vm->network_writeable(ch);
}

/// @note called by main thread
void EMU_OSD::socket_readable(int ch)
{
	// network is readable
	logging->out_debugf(_T("readable. ch=%d"), ch);
	if (socs[ch].device) socs[ch].device->network_readable(ch);
//	else vm->network_readable(ch);
}

static int memcpy_remove_dup_iac(char *dst, const char *src, int size)
{
	int pos=0;
	for(int i=0; i<size;) {
		if (src[i] == (char)0xff && i+1 == size) {
			break;
		}
		dst[pos++]=src[i];
		if (src[i] == (char)0xff && src[i+1] == (char)0xff) {
			i++;
		}
		i++;
	}
	return pos;
}

/// @note called by main thread
void EMU_OSD::check_socket()
{
	// check arrived packets
	if (socket_nums > 0) {
		int retval = SDLNet_CheckSockets(socket_set, 0);
		if (retval < 0) {
			// error
			logging->out_logf(LOG_ERROR, _T("SDLNet_CheckSockets: %s"), SDLNet_GetError());
		} else if (retval > 0) {
			// data arrived
			for(int i = 0; i < SOCKET_MAX; i++) {
				struct st_soc *soc = &socs[i];
				if (soc->is_tcp) {
					if (SDLNet_SocketReady(soc->tcp_socket)) {
						if (soc->is_server) {
							socket_accept(i);
						} else {
							recv_data(i);
							socket_readable(i);
						}
					}
				} else {
					if (SDLNet_SocketReady(soc->udp_socket)) {
						recv_data(i);
						socket_readable(i);
					}
				}
			}
		}
	}
}

/// @note called by emu thread
void EMU_OSD::update_socket()
{
	// buffer copy
	for(int i = 0; i < SOCKET_MAX; i++) {
		struct st_soc *soc = &socs[i];
		if(soc->recv_r_ptr < soc->recv_w_ptr) {
			if (soc->device) {
				// get buffer
				int size0 = 0;
				int size1 = 0;
				int flags = 0;
				uint8_t* buf0 = soc->device->get_recvbuffer0(i, &size0, &size1, &flags);
				//	buf0 = vm->get_recvbuffer0(i, &size0, &size1, &flags);
				uint8_t* buf1 = soc->device->get_recvbuffer1(i);
				//	buf1 = vm->get_recvbuffer1(i);

				int size = soc->recv_w_ptr - soc->recv_r_ptr;
				if(size > size0 + size1) {
					size = size0 + size1;
				}
				char* src = &soc->recv_buffer[soc->recv_r_ptr];
				soc->recv_r_ptr += size;

				if(size <= size0) {
					if (!(flags & 1)) {
						memcpy(buf0, src, size);
					} else {
						// escape telnet iac
						size = memcpy_remove_dup_iac((char *)buf0, src, size);
					}
				}
				else {
					if (!(flags & 1)) {
						memcpy(buf0, src, size0);
						memcpy(buf1, src + size0, size - size0);
					} else {
						// escape telnet iac
						size0 = memcpy_remove_dup_iac((char *)buf0, src, size0);
						size = memcpy_remove_dup_iac((char *)buf1, src + size0, size - size0);
					}
				}
				soc->device->inc_recvbuffer_ptr(i, size);
				//	vm->inc_recvbuffer_ptr(i, size);
			}
		}
		else if(soc->socket_delay != 0) {
			if(--soc->socket_delay == 0) {
				disconnect_socket(i);
//				if (soc->device) soc->device->network_disconnected(i);
//				else vm->network_disconnected(i);
			}
		}
	}
}

/// @note called by main thread
void EMU_OSD::socket_accept(int ch)
{
	if (ch < 0 || SOCKET_MAX <= ch) return;

	int new_ch = -1;

	TCPsocket new_tcpsock;

	new_tcpsock=SDLNet_TCP_Accept(socs[ch].tcp_socket);
	if (new_tcpsock == NULL) {
		logging->out_logf(LOG_ERROR, _T("SDLNet_TCP_Accept: ch=%d %s"), ch, SDLNet_GetError());
		return;
	}

	// find empty socket
	new_ch = get_socket_channel();
	if (new_ch < 0) {
		SDLNet_TCP_Close(new_tcpsock);
		return;
	}

	int numused = SDLNet_TCP_AddSocket(socket_set, new_tcpsock);
	if(numused == -1) {
		SDLNet_TCP_Close(new_tcpsock);
		logging->out_logf(LOG_ERROR, _T("EMU_OSD::socket_accept: SDLNet_TCP_AddSocket: new_ch=%d %s"), new_ch, SDLNet_GetError());
		return;
	}

	IPaddress *remote_ip = SDLNet_TCP_GetPeerAddress(new_tcpsock);
	if (remote_ip) {
		logging->out_debugf(_T("accepted from 0x%08x"), remote_ip->host);
	}

	socs[new_ch].tcp_socket = new_tcpsock;
	socs[new_ch].is_tcp = socs[ch].is_tcp;
	socs[new_ch].device = socs[ch].device;
	socs[new_ch].mux = new CMutex();
	socket_nums++;

	if (socs[ch].device) socs[ch].device->network_accepted(ch, new_ch);
//	else vm->network_accepted(ch, new_ch);

	logging->out_debugf(_T("EMU_OSD::socket_accept: ok. ch=%d is_tcp=%s new_ch=%d is_tcp=%s")
		, ch, socs[ch].is_tcp ? _T("true") : _T("false"), new_ch, socs[new_ch].is_tcp ? _T("true") : _T("false"));

	// TODO: post SDL event
	// note that i'll fire the writeable event only, won't fire the connected event.
#ifndef USE_SDL_EVENT
	st_timer *timer2 = new st_timer;
	timer2->ch = ch;
	timer2->id = SDL_AddTimer(32, on_writeable, (void *)timer2);
#else
#ifdef USE_GTK
	g_signal_emit_by_name(G_OBJECT(window), "user-socket"
		, new_ch << 16 | SDL_USEREVENT_SOCKET_WRITEABLE);
#else
	SDL_Event evt;
	evt.type = SDL_USEREVENT_SOCKET;
	evt.user.code = SDL_USEREVENT_SOCKET_WRITEABLE;
	evt.user.data1 = (void *)(intptr_t)new_ch;
	evt.user.data2 = NULL;
	SDL_PushEvent(&evt);
#endif
#endif
	return;
}

/// @note called by main thread
bool EMU_OSD::init_socket_tcp(int ch, DEVICE *dev, bool server)
{
	if (ch < 0 || SOCKET_MAX <= ch) {
		logging->out_logf(LOG_ERROR, _T("EMU_OSD::init_socket_tcp: invalid socket channel. ch=%d"), ch);
		return false;
	}

	struct st_soc *soc = &socs[ch];
	disconnect_socket(ch);

	// reserve channel for tcp connection
	soc->is_tcp = true;
	soc->device = NULL;

	soc->recv_r_ptr = soc->recv_w_ptr = 0;
	soc->device = dev;
	return true;
}

/// @note called by main thread
bool EMU_OSD::init_socket_udp(int ch, DEVICE *dev, bool server)
{
	if (ch < 0 || SOCKET_MAX <= ch) {
		logging->out_logf(LOG_ERROR, _T("EMU_OSD::init_socket_udp: invalid socket channel. ch=%d"), ch);
		return false;
	}

	struct st_soc *soc = &socs[ch];

	disconnect_socket(ch);

	// reserve channel for udp connection
	soc->is_tcp = false;
	soc->device = NULL;

	if (!server) {
		// client
		UDPsocket sock;

		sock = SDLNet_UDP_Open(0);
		if(!sock) {
			logging->out_logf(LOG_ERROR, _T("SDLNet_UDP_Open: ch=%d %s"), ch, SDLNet_GetError());
			return false;
		}
		int numused = SDLNet_UDP_AddSocket(socket_set, sock);
		if(numused == -1) {
			SDLNet_UDP_Close(sock);
			logging->out_logf(LOG_ERROR, _T("SDLNet_UDP_AddSocket: ch=%d %s"), ch, SDLNet_GetError());
			return false;
		}
		soc->udp_socket = sock;
		soc->is_server = server;
	}

	soc->recv_r_ptr = soc->recv_w_ptr = 0;
	soc->device = dev;
	return true;
}

bool EMU_OSD::connect_socket(int ch, IPaddress *ip, bool server)
{
	if (ch < 0 || SOCKET_MAX <= ch) {
		logging->out_logf(LOG_ERROR, _T("EMU_OSD::connect_socket: invalid socket channel. ch=%d"), ch);
		return false;
	}

	struct st_soc *soc = &socs[ch];

	if (soc->is_tcp) {
		TCPsocket sock;

		if (server == false && (ip->host == INADDR_NONE || ip->host == INADDR_ANY)) {
			logging->out_log(LOG_ERROR, _T("Cannot create socket as the server."));
			return false;
		}

		sock = SDLNet_TCP_Open(ip, server ? SDL_TRUE : SDL_FALSE);
		if(!sock) {
			logging->out_logf(LOG_ERROR, _T("SDLNet_TCP_Open: ch=%d %s"), ch, SDLNet_GetError());
			return false;
		}
		int numused = SDLNet_TCP_AddSocket(socket_set, sock);
		if(numused == -1) {
			SDLNet_TCP_Close(sock);
			logging->out_logf(LOG_ERROR, _T("SDLNet_TCP_AddSocket: ch=%d %s"), ch, SDLNet_GetError());
			return false;
		}
		soc->tcp_socket = sock;
	} else {
		if (server) {
			UDPsocket sock;

			sock = SDLNet_UDP_Open(ip->port);
			if(!sock) {
				logging->out_logf(LOG_ERROR, _T("SDLNet_UDP_Open: ch=%d %s"), ch, SDLNet_GetError());
				return false;
			}
			int channel = SDLNet_UDP_Bind(sock, -1, ip);
			if (channel == -1) {
				SDLNet_UDP_Close(sock);
				logging->out_logf(LOG_ERROR, _T("SDLNet_UDP_Bind: ch=%d %s"), ch, SDLNet_GetError());
				return false;
			}
			int numused = SDLNet_UDP_AddSocket(socket_set, sock);
			if(numused == -1) {
				SDLNet_UDP_Close(sock);
				logging->out_logf(LOG_ERROR, _T("SDLNet_UDP_AddSocket: ch=%d %s"), ch, SDLNet_GetError());
				return false;
			}
			soc->udp_socket = sock;
		}
	}
	if (!server) soc->mux = new CMutex();
	socket_nums++;
	soc->is_server = server;

	// TODO: post SDL event
	if (!server) {
#ifndef USE_SDL_EVENT
		st_timer *timer1 = new st_timer;
		timer1->ch = ch;
		timer1->id = SDL_AddTimer(16, on_connected, (void *)timer1);
		st_timer *timer2 = new st_timer;
		timer2->ch = ch;
		timer2->id = SDL_AddTimer(32, on_writeable, (void *)timer2);
#else
#ifdef USE_GTK
		g_signal_emit_by_name(G_OBJECT(window), "user-socket"
			, ch << 16 | SDL_USEREVENT_SOCKET_CONNECTED);

		g_signal_emit_by_name(G_OBJECT(window), "user-socket"
			, ch << 16 | SDL_USEREVENT_SOCKET_WRITEABLE);
#else
		SDL_Event evt;
		evt.type = SDL_USEREVENT_SOCKET;
		evt.user.code = SDL_USEREVENT_SOCKET_CONNECTED;
		evt.user.data1 = (void *)(intptr_t)ch;
		evt.user.data2 = NULL;
		SDL_PushEvent(&evt);

		evt.type = SDL_USEREVENT_SOCKET;
		evt.user.code = SDL_USEREVENT_SOCKET_WRITEABLE;
		evt.user.data1 = (void *)(intptr_t)ch;
		evt.user.data2 = NULL;
		SDL_PushEvent(&evt);
#endif
#endif
	}
	return true;
}

bool EMU_OSD::connect_socket(int ch, uint32_t ipaddr, int port, bool server)
{
	IPaddress ip;

	ip.host = (Uint32)ipaddr;
	ip.port = (Uint16)port;

	return connect_socket(ch, &ip, server);
}

bool EMU_OSD::connect_socket(int ch, const _TCHAR *hostname, int port, bool server)
{
	IPaddress ip;

	if(!get_ipaddr_(hostname, port, &ip)) {
		return false;
	}

	return connect_socket(ch, &ip, server);
}

bool EMU_OSD::get_ipaddr(const _TCHAR *hostname, int port, uint32_t *ip)
{
	IPaddress nip;
	bool rc = get_ipaddr_(hostname, port, &nip);
	if (ip) *ip = nip.host;
	return rc;
}

bool EMU_OSD::get_ipaddr_(const _TCHAR *hostname, int port, IPaddress *ip)
{
#ifdef _UNICODE
	char host[_MAX_PATH];
	wcstombs(host, hostname, _MAX_PATH);
#else
	const char *host = hostname;
#endif

	if(SDLNet_ResolveHost(ip, (const char *)host, (Uint16)port) == -1) {
		// unresolved hostname to ipaddr
		logging->out_logf(LOG_ERROR, _T("SDLNet_ResolveHost: %s"), SDLNet_GetError());
		return false;
	}

	return true;
}

void EMU_OSD::disconnect_socket(int ch)
{
	if (ch < 0 || SOCKET_MAX <= ch) return;

	struct st_soc *soc = &socs[ch];

	if (soc->is_tcp) {
		if (soc->tcp_socket) {
			SDLNet_TCP_Close(soc->tcp_socket);
			SDLNet_TCP_DelSocket(socket_set, soc->tcp_socket);
			soc->tcp_socket = NULL;
			soc->socket_delay = 0;
			socket_nums--;
			if (soc->device) soc->device->network_disconnected(ch);
//			else vm->network_disconnected(ch);
		}
	} else {
		if (soc->udp_socket) {
			SDLNet_UDP_Close(soc->udp_socket);
			SDLNet_UDP_DelSocket(socket_set, soc->udp_socket);
			soc->udp_socket = NULL;
			soc->socket_delay = 0;
			socket_nums--;
			if (soc->device) soc->device->network_disconnected(ch);
//			else vm->network_disconnected(ch);
		}
	}

	soc->device = NULL;

	delete soc->mux;
	soc->mux = NULL;

	// TODO: post SDL event
#ifndef USE_SDL_EVENT
	st_timer *timer = new st_timer;
	timer->ch = ch;
	timer->id = SDL_AddTimer(16, on_disconnected, (void *)timer);
#else
#ifdef USE_GTK
	g_signal_emit_by_name(G_OBJECT(window), "user-socket"
		, ch << 16 | SDL_USEREVENT_SOCKET_DISCONNECTED);
#else
	SDL_Event evt;
	evt.type = SDL_USEREVENT_SOCKET;
	evt.user.code = SDL_USEREVENT_SOCKET_DISCONNECTED;
	evt.user.data1 = (void *)(intptr_t)ch;
	evt.user.data2 = NULL;
	SDL_PushEvent(&evt);
#endif
#endif
}


int EMU_OSD::get_socket_channel()
{
	int new_ch = -1;
	for(int i = 0; i < SOCKET_MAX; i++) {
		if (socs[i].tcp_socket == NULL && socs[i].udp_socket == NULL) {
			new_ch = i;
			break;
		}
	}
	if (new_ch < 0) {
		logging->out_log(LOG_ERROR, _T("Socket is full."));
	}
	return new_ch;
}

bool EMU_OSD::listen_socket(int ch)
{
	return false;
}

/// @note called by emu thread
void EMU_OSD::send_data_tcp(int ch)
{
	if (ch < 0 || SOCKET_MAX <= ch) return;

	if(socs[ch].is_tcp) {
		send_data(ch);
	}
}

/// @note called by emu thread
void EMU_OSD::send_data_udp(int ch, uint32_t ipaddr, int port)
{
	if (ch < 0 || SOCKET_MAX <= ch) return;

	struct st_soc *soc = &socs[ch];

	if(!soc->is_tcp) {
		udp_packet.address.host = (Uint32)ipaddr;
		udp_packet.address.port = (Uint16)port;
		udp_packet.channel = -1;

		send_data(ch);
	}
}

/// @note called by emu thread
void EMU_OSD::send_data_udp(int ch, const _TCHAR *hostname, int port)
{
	IPaddress ip;

	if (!get_ipaddr_(hostname, port, &ip)) {
		return;
	}

	send_data_udp(ch, (uint32_t)ip.host, ip.port);
}

/// @note usually called by emu thread
/// @note called by main thread at first when connection succeeded.
void EMU_OSD::send_data(int ch)
{
	if (ch < 0 || SOCKET_MAX <= ch) return;

	struct st_soc *soc = &socs[ch];

	soc->mux->lock();

	// get send buffer and data size
	int size;
	int flags = 0;
	uint8_t* buf;
	uint8_t bufcnvd[1024];

	// loop while send buffer is not empty
	while(soc->device) {
		buf = soc->device->get_sendbuffer(ch, &size, &flags);
		//	buf = vm->get_sendbuffer(ch, &size);

		if(!buf || !size) {
			break;
		}
		if (flags & 1) {
			// escape telnet iac
			int pos=0;
			for(int i=0; i<size && pos<1023; i++) {
				bufcnvd[pos++]=buf[i];
				if (buf[i] == 0xff) {
					bufcnvd[pos++]=buf[i];
				}
			}
			size=pos;
			buf=bufcnvd;
		}
		if(soc->is_tcp) {
			int sent_size = SDLNet_TCP_Send(soc->tcp_socket, (const void *)buf, size);
			if(sent_size < size) {
				// error
				disconnect_socket(ch);
				socket_disconnected(ch);
				break;
			}
			size = sent_size;
		} else {
			udp_packet.data = (Uint8 *)buf;
			udp_packet.maxlen = size;
			udp_packet.len = size;
			int sent_packet = SDLNet_UDP_Send(soc->udp_socket, udp_packet.channel, &udp_packet);
			if(sent_packet <= 0) {
				// error
				disconnect_socket(ch);
				socket_disconnected(ch);
				break;
			}
		}
		soc->device->inc_sendbuffer_ptr(ch, size);
		//	vm->inc_sendbuffer_ptr(ch, size);
	}

	soc->mux->unlock();
}

/// Receive data from socket, and store to first buffer to be able to read from emu devices.
/// @note called by main thread
void EMU_OSD::recv_data(int ch)
{
	if (ch < 0 || SOCKET_MAX <= ch) return;
	struct st_soc *soc = &socs[ch];

	if (soc->recv_buffer == NULL) {
		soc->recv_buffer = new char[SOCKET_BUFFER_MAX];
	}

	if(soc->is_tcp) {
		int size = SOCKET_BUFFER_MAX - soc->recv_w_ptr;
		char* buf = &soc->recv_buffer[soc->recv_w_ptr];
		int recv_size = SDLNet_TCP_Recv(soc->tcp_socket, (void *)buf, size);
		if(recv_size <= 0) {
			disconnect_socket(ch);
			socket_disconnected(ch);
			return;
		}
		soc->recv_w_ptr += recv_size;
	}
	else {
		UDPpacket packet;
		int size = SOCKET_BUFFER_MAX - soc->recv_w_ptr;
		char* buf = &soc->recv_buffer[soc->recv_w_ptr];

		packet.address.host = 0;
		packet.address.port = 0;
		packet.channel = 0;
		packet.maxlen = size - 8;
		packet.len = 0;
		packet.data = (Uint8 *)(buf + 8);

		if(size < 8) {
			return;
		}
		int recv_packet = SDLNet_UDP_Recv(soc->udp_socket, &packet);
//		if((size = recvfrom(soc->soc, buf + 8, size - 8, 0, (struct sockaddr *)&addr, &len)) == SOCKET_ERROR) {
		if(recv_packet < 0) {
			disconnect_socket(ch);
			socket_disconnected(ch);
			return;
		}
//		size += 8;
		buf[0] = packet.len >> 8;
		buf[1] = packet.len;
		buf[2] = (char)packet.address.host;
		buf[3] = (char)(packet.address.host >> 8);
		buf[4] = (char)(packet.address.host >> 16);
		buf[5] = (char)(packet.address.host >> 24);
		buf[6] = (char)packet.address.port;
		buf[7] = (char)(packet.address.port >> 8);
		soc->recv_w_ptr += packet.len;
	}
}

const void *EMU_OSD::get_socket(int ch) const
{
	if (socs[ch].is_tcp) {
		return socs[ch].tcp_socket;
	} else {
		return socs[ch].udp_socket;
	}
}
#endif /* USE_SOCKET */

#ifndef USE_SDL_EVENT
//
// callback on timer
//
static Uint32 on_connected(Uint32 interval, void *param)
{
	st_timer *timer = (st_timer *)param;
	emu->socket_connected(timer->ch);
	SDL_RemoveTimer(timer->id);
	delete timer;
	return interval;
}
static Uint32 on_disconnected(Uint32 interval, void *param)
{
	st_timer *timer = (st_timer *)param;
	emu->socket_disconnected(timer->ch);
	SDL_RemoveTimer(timer->id);
	delete timer;
	return interval;
}
static Uint32 on_writeable(Uint32 interval, void *param)
{
	st_timer *timer = (st_timer *)param;
	emu->socket_writeable(timer->ch);
	emu->send_data(timer->ch);
	SDL_RemoveTimer(timer->id);
	delete timer;
	return interval;
}
static Uint32 on_readable(Uint32 interval, void *param)
{
	st_timer *timer = (st_timer *)param;
	emu->recv_data(timer->ch);
	emu->socket_readable(timer->ch);
	SDL_RemoveTimer(timer->id);
	delete timer;
	return interval;
}
#endif // USE_SDL_EVENT

#ifdef USE_GTK
void OnUserSocket(GtkWidget *widget, gint data, gpointer user_data)
{
	EMU *emu = (EMU *)user_data;
	int no = (data & 0xffff0000) >> 16;
	int code = (data & 0xffff);
	switch(code) {
		case SDL_USEREVENT_SOCKET_CONNECTED:
			emu->socket_connected(no);
			break;
		case SDL_USEREVENT_SOCKET_DISCONNECTED:
			emu->socket_disconnected(no);
			break;
		case SDL_USEREVENT_SOCKET_WRITEABLE:
			emu->socket_writeable(no);
			emu->send_data(no);
			break;
//		case SDL_USEREVENT_SOCKET_READABLE:
//			emu->recv_data(no);
//			emu->socket_readable(no);
//			break;
//		case SDL_USEREVENT_SOCKET_ACCEPT:
//			emu->socket_accept(no);
//			break;
	}
}
#endif
