/** @file emu_screen.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2017.12.01

	@brief [ emu screen ]
*/

#include "../emu.h"
#include "../depend.h"
#include "../config.h"
#include "../cpixfmt.h"
#include "../csurface.h"
#include "../video/rec_video.h"
#include "../cmutex.h"


void EMU::EMU_SCREEN()
{
	RECT_IN(screen_size, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
	SIZE_IN(screen_aspect_size, SCREEN_WIDTH_ASPECT, SCREEN_HEIGHT_ASPECT);
//	window_width = WINDOW_WIDTH;
//	window_height = WINDOW_HEIGHT;
//	SIZE_IN(min_window_size, MIN_WINDOW_WIDTH, MIN_WINDOW_HEIGHT);
	screen_size_changed = true;
	SIZE_IN(desktop_size, MIN_WINDOW_WIDTH, MIN_WINDOW_HEIGHT);
	SIZE_IN(display_size, MIN_WINDOW_WIDTH, MIN_WINDOW_HEIGHT);
	desktop_bpp = 24;
	display_margin.left = 0;
	display_margin.top = 0;
	display_margin.right = 0;
	display_margin.bottom = 0;

	source_size = screen_size;
	SIZE_IN(source_aspect_size, -1, -1);
#ifdef USE_SCREEN_ROTATE
	// clockwise rotate
	MATRIX_IN(rotate_matrix[0],  1,  0,  0,  1);
	MATRIX_IN(rotate_matrix[1],  0, -1,  1,  0);
	MATRIX_IN(rotate_matrix[2], -1,  0,  0, -1);
	MATRIX_IN(rotate_matrix[3],  0,  1, -1,  0);
#endif
#ifdef USE_SMOOTH_STRETCH
	SIZE_IN(stretch_power, -1, -1);
	stretch_screen = false;
#endif
	RECT_IN(stretched_size, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
	POINT_IN(stretched_dest_real, 0, 0);
	mixed_size = source_size;
	SIZE_IN(mixed_ratio, 1, 1);

	// screen mode
	window_mode_power = 10;
	now_fullscreen = false;
	prev_window_mode = 0;
	window_dest.x = 0;
	window_dest.y = 0;
	first_change_screen = true;

	// initialize video recording
	rec_video_size[0].w = MIN_WINDOW_WIDTH;
	rec_video_size[0].h = MIN_WINDOW_HEIGHT;
	rec_video_size[0].x = (MAX_WINDOW_WIDTH - MIN_WINDOW_WIDTH) / 2;
	rec_video_size[0].y = (MAX_WINDOW_HEIGHT - MIN_WINDOW_HEIGHT) / 2;
	rec_video_size[1].w = MAX_WINDOW_WIDTH;
	rec_video_size[1].h = MAX_WINDOW_HEIGHT;
	rec_video_size[1].x = 0;
	rec_video_size[1].y = 0;
	rec_video_stretched_size = rec_video_size[0];

	rec_video = new REC_VIDEO(this);
	now_recording_video = rec_video->GetNowRecordingPtr();

	// initialize update flags
	disable_screen = DISABLE_SURFACE;
	first_invalidate = self_invalidate = false;
	skip_frame = false;

	//
	sufOrigin = new CSurface();
#ifdef USE_SCREEN_ROTATE
	sufRotate = new CSurface();
#endif
	sufSource = NULL;
	sufMixed = new CSurface();

#ifdef USE_SMOOTH_STRETCH
	sufStretch1 = new CSurface();
	sufStretch2 = new CSurface();
#endif

	pixel_format = new CPixelFormat();

	// for lock screen
	mux_update_screen = new CMutex();
	screen_changing = false;
}

///
/// initialize screen
///
void EMU::initialize_screen()
{
	logging->out_debug(_T("EMU::initialize_screen"));

	change_pixel_aspect(config.pixel_aspect);
	change_rec_video_size(config.screen_video_size);
}

///
/// release screen
///
void EMU::release_screen()
{
	delete sufOrigin;
	sufOrigin = NULL;
#ifdef USE_SCREEN_ROTATE
	delete sufRotate;
	sufRotate = NULL;
#endif
#ifdef USE_EMU_INHERENT_SPEC
	delete sufMixed;
	sufMixed = NULL;
#endif
#ifdef USE_SMOOTH_STRETCH
	delete sufStretch1;
	sufStretch1 = NULL;
	delete sufStretch2;
	sufStretch2 = NULL;
#endif

	delete pixel_format;
	pixel_format = NULL;

	delete mux_update_screen;
	mux_update_screen = NULL;

	delete rec_video;
	rec_video = NULL;
}

///
/// get RGB format in a pixeldata
///
void EMU::get_rgbformat(uint32_t *r_mask, uint32_t *g_mask, uint32_t *b_mask, uint8_t *r_shift, uint8_t *g_shift, uint8_t *b_shift)
{
	pixel_format->Get(r_mask, g_mask, b_mask, r_shift, g_shift, b_shift);
}

///
/// get RGB + alpha format in a pixeldata
///
void EMU::get_rgbaformat(uint32_t *r_mask, uint32_t *g_mask, uint32_t *b_mask, uint32_t *a_mask, uint8_t *r_shift, uint8_t *g_shift, uint8_t *b_shift, uint8_t *a_shift)
{
	pixel_format->Get(r_mask, g_mask, b_mask, a_mask, r_shift, g_shift, b_shift, a_shift);
}

///
/// get R,G,B in a pixel
///
void EMU::get_rgbcolor(scrntype pixel, uint8_t *r, uint8_t *g, uint8_t *b)
{
	pixel_format->Get(pixel, *r, *g, *b);
}

///
/// create a pixeldata from R,G,B 
///
scrntype EMU::map_rgbcolor(uint8_t r, uint8_t g, uint8_t b)
{
	return pixel_format->Map(r, g, b);
}

void EMU::release_screen_on_emu_thread()
{
	// stop video recording
	stop_rec_video();
}

void EMU::set_display_size(int width, int height, int power, bool window_mode)
{
}

/// set vm drawing area
void EMU::set_vm_display_size()
{
	if (vm) {
		int new_size = config.screen_video_size;
		int l = mixed_size.x;
		int t = mixed_size.y;
		int r = l + mixed_size.w;
		int b = t + mixed_size.h;

		// vm drawing area is same as area to capture and record screen
		if (rec_video_size[new_size].w > mixed_size.w) {
			l = rec_video_size[new_size].x;
			r = rec_video_size[new_size].x + rec_video_size[new_size].w;
		}

		if (rec_video_size[new_size].h > mixed_size.h) {
			t = rec_video_size[new_size].y;
			b = rec_video_size[new_size].y + rec_video_size[new_size].h;
		}

#ifdef USE_SCREEN_ROTATE
		l = screen_size.x;
		t = screen_size.y;
		r = l + screen_size.w;
		b = t + screen_size.h;
#endif
		vm->set_display_size(l,t,r,b);
	}
}

/// reserve top margin area for ledbox
int EMU::adjust_y_position(int h, int y)
{
	if (20 < h - LIMIT_MIN_WINDOW_HEIGHT && h - LIMIT_MIN_WINDOW_HEIGHT < 42) {
		y -= (42 - h + LIMIT_MIN_WINDOW_HEIGHT) / 2;
	}
	return y;
}

void EMU::draw_screen()
{
}

bool EMU::mix_screen()
{
	return false;
}

void EMU::fill_gray()
{
	scrntype color = map_rgbcolor(0x40, 0x40, 0x40);
	for(int y=0; y<source_size.h; y++) {
		scrntype *line = screen_buffer(y);
		for(int x=0; x<source_size.w; x++) {
			*line = color;
			line++;
		}
	}
}

scrntype* EMU::screen_buffer(int y)
{
	return NULL;
}

int EMU::screen_buffer_offset()
{
	return 0;
}

bool EMU::now_skip_frame()
{
	return skip_frame && !now_rec_video();
}

void EMU::capture_screen()
{
}

///
/// start recording video
///
bool EMU::start_rec_video(int type, int fps_no, bool show_dialog)
{
#ifdef USE_REC_VIDEO
	int size = config.screen_video_size;
	return rec_video->Start(type, fps_no, rec_video_size[size], sufOrigin, show_dialog);
#else
	return false;
#endif
}

///
/// record video
///
void EMU::record_rec_video()
{
}

///
/// stop recording video
///
void EMU::stop_rec_video()
{
#ifdef USE_REC_VIDEO
	rec_video->Stop();
#endif
}

///
/// restart recording video
///
void EMU::restart_rec_video()
{
#ifdef USE_REC_VIDEO
	rec_video->Restart();
#endif
}

/// fps number
/// @return 0 - 5 or -1
int EMU::get_rec_video_fps_num()
{
#ifdef USE_REC_VIDEO
	return rec_video->GetFpsNum();
#else
	return -1;
#endif
}

bool EMU::now_rec_video()
{
#ifdef USE_REC_VIDEO
	return *now_recording_video;
#else
	return false;
#endif
}

void EMU::resize_rec_video(int num)
{
#ifdef USE_EMU_INHERENT_SPEC
	config.screen_video_size = num;
	change_rec_video_size(num);

	// send display size to vm
	set_vm_display_size();
#endif
}
void EMU::change_rec_video_size(int num)
{
	rec_video_stretched_size = rec_video_size[num];
	if (mixed_ratio.w < mixed_ratio.h) {
		rec_video_stretched_size.h = rec_video_stretched_size.h * mixed_ratio.w / mixed_ratio.h;
		rec_video_stretched_size.y += (rec_video_size[num].h - rec_video_stretched_size.h) / 2; 
	} else {
		rec_video_stretched_size.w = rec_video_stretched_size.w * mixed_ratio.h / mixed_ratio.w;
		rec_video_stretched_size.x += (rec_video_size[num].w - rec_video_stretched_size.w) / 2; 
	}
	rec_video_stretched_size.y = adjust_y_position(rec_video_stretched_size.h, rec_video_stretched_size.y);

#ifdef _DEBUG_LOG
	logging->out_debugf(_T("          rec_video_size: num:%d x:%d y:%d w:%d h:%d")
		,num,rec_video_size[num].x,rec_video_size[num].y,rec_video_size[num].w,rec_video_size[num].h);
	logging->out_debugf(_T("rec_video_stretched_size:       x:%d y:%d w:%d h:%d")
		,rec_video_stretched_size.x,rec_video_stretched_size.y,rec_video_stretched_size.w,rec_video_stretched_size.h);
#endif
}

bool EMU::rec_video_enabled(int type)
{
#ifdef USE_REC_VIDEO
	return rec_video->IsEnabled(type);
#else
	return false;
#endif
}

const _TCHAR **EMU::get_rec_video_codec_list(int type)
{
#ifdef USE_REC_VIDEO
	return rec_video->GetCodecList(type);
#else
	return NULL;
#endif
}

const CMsg::Id *EMU::get_rec_video_quality_list(int type)
{
#ifdef USE_REC_VIDEO
	return rec_video->GetQualityList(type);
#else
	return NULL;
#endif
}

//----------

/// enumerate screen mode for window
/// calcurate magnify range
/// @param [in] max_width
/// @param [in] max_height
void EMU::enum_window_mode(int max_width, int max_height)
{
	window_mode.Enum(max_width, max_height);
}


/// matching window or fullscreen size
void EMU::find_screen_mode()
{
	if (config.window_mode < 8) {
		// matching window size at last time
		int find = window_mode.Find(config.screen_width, config.screen_height);
		if (find >= 0) {
			config.window_mode = find;
		} else {
			config.window_mode = 0;
		}
	} else {
		// matching fullscreen size at last time
		int find = screen_mode.FindMode(config.disp_device_no, config.screen_width, config.screen_height);
		if (find >= 0) {
			config.window_mode = find + 8;
		} else {
			config.window_mode = 0;
		}
	}
}

void EMU::enum_screen_mode(uint32_t flags)
{
	screen_mode.Enum(desktop_size.w, desktop_size.h, flags);
}

/// change window size / switch over fullscreen and window
/// @param[in] mode 0 - 7: window size  8 -:  fullscreen size  -1: switch over  -2: shift window mode
void EMU::change_screen_mode(int mode)
{
//	logging->out_debugf(_T("change_screen_mode: mode:%d cwmode:%d pwmode:%d w:%d h:%d"),mode,config.window_mode,prev_window_mode,desktop_size.w,desktop_size.h);
//	if (mode == config.window_mode) return;

	if (mode == -1) {
		// switch over fullscreen and window
		if (now_fullscreen) {
			// go window mode
			mode = prev_window_mode;
		}
	} else if (mode == -2) {
		// shift window mode 
		if (now_fullscreen) {
			// no change
			return;
		} else {
			mode = ((config.window_mode + 1) % window_mode.Count());
		}
	}
	if (!now_fullscreen) {
		prev_window_mode = config.window_mode;
	}
//	logging->out_debugf(_T("change_screen_mode: mode:%d cwmode:%d pwmode:%d w:%d h:%d"),mode,config.window_mode,prev_window_mode,desktop_size.w,desktop_size.h);
	set_window(mode, desktop_size.w, desktop_size.h);
}

/// change screen stretching on fullscreen mode
///
/// @param[in] num : 0:no expand  1:normal expand  2:maximum expand
///
/// @attention should be called by main thread
void EMU::change_stretch_screen(int num)
{
	if (num < 0) {
		num = (config.stretch_screen + 1);
	}
	num = (num % 3);
	if (config.stretch_screen == num) {
		num = 0;
	}
	config.stretch_screen = num;
	if (now_fullscreen) {
		// change screen
		set_display_size(-1, -1, 10, !now_fullscreen);
	}
}

void EMU::set_window(int mode, int cur_width, int cur_height)
{
}

/// change aspect ratio
///
/// @param[in] mode : aspect 0:square 1: 2:
void EMU::change_pixel_aspect(int mode)
{
	if (mode < 0) {
		mode = (config.pixel_aspect + 1) % get_pixel_aspect_count();
	}
	config.pixel_aspect = get_pixel_aspect(mode, &mixed_ratio.w, &mixed_ratio.h);

	set_display_size(config.screen_width, config.screen_height, now_fullscreen ? 10 : window_mode_power, !now_fullscreen);
}

int EMU::get_pixel_aspect_count()
{
	return 3;
}

int EMU::get_pixel_aspect(int mode, int *wratio, int *hratio)
{
	int new_mode = mode;
	switch(mode) {
	case 1:
		*wratio = 10;
		*hratio = 11;
		break;
	case 2:
		*wratio = 10;
		*hratio = 12;
		break;
	default:
		*wratio = 1;
		*hratio = 1;
		new_mode = 0;
		break;
	}
#ifdef USE_SCREEN_ROTATE
	if (config.monitor_type & 1) {
		SWAP(int, *wratio, *hratio);
	}
#endif
	return new_mode;
}

int EMU::get_window_mode_count() const
{
	return window_mode.Count();
}

const window_mode_t *EMU::get_window_mode(int num) const
{
	return window_mode.Get(num);
}

int EMU::get_screen_mode_count() const
{
	return screen_mode.CountMode();
}

const screen_mode_t *EMU::get_screen_mode(int num) const
{
	return screen_mode.GetMode(num);
}

VmRectWH *EMU::get_screen_record_size(int num)
{
	if (0 <= num && num <= 1) {
		return &rec_video_size[num];
	}
	return NULL;
}

void EMU::lock_screen()
{
	mux_update_screen->lock();
	screen_changing = true;
}

void EMU::unlock_screen()
{
	screen_changing = false;
	mux_update_screen->unlock();
}

